# 1) 데이터 불러오기
import tensorflow as tf
from tensorflow import keras
import numpy as np
import matplotlib.pyplot as plt

print(tf.__version__)   # Tensorflow의 버전 출력
mnist = keras.datasets.mnist
# MNIST 데이터를 로드
(x_train, y_train), (x_test, y_test) = mnist.load_data()   

print(len(x_train))  # x_train 배열의 크기 출력

# 2) 데이터 확인하기
plt.imshow(x_train[1], cmap=plt.cm.binary) #color map. binary는 값이 작을수록 검은색, 클수록 흰색으로 표시
plt.show()

print(y_train[1])

# 
index=100
plt.imshow(x_train[index], cmap=plt.cm.binary)
plt.show()
print((index+1),'번째 이미지는 ', y_train[index], '입니다.')

# %%
print(x_train.shape) #(60000, 28, 28)은 6000개의 sample set, 28x28 픽셀 크기의 2차원 배열
print(x_test.shape)

# 3) 데이터 전처리
print('최소값: ',np.min(x_train), ' 최대값: ',np.max(x_train)) #픽셀 값이 0~255 범위임

# %%
x_train_norm, x_test_norm = x_train/255.0, x_test/255.0
print('최소값: ', np.min(x_train_norm), '최댓값: ',np.max(x_train_norm)) #데이터 전처리

# 4) 딥러닝 네트워크 설계하기
model=keras.models.Sequential()
model.add(keras.layers.Conv2D(16, (3,3), activation='relu', input_shape=(28,28,1)))
model.add(keras.layers.MaxPool2D(2,2))
model.add(keras.layers.Conv2D(32, (3,3), activation='relu'))
model.add(keras.layers.MaxPooling2D((2,2)))
model.add(keras.layers.Flatten())
model.add(keras.layers.Dense(32, activation='relu'))
model.add(keras.layers.Dense(10, activation='softmax'))

print('Model에 추가된 Layer 개수: ', len(model.layers))

"""
<이미지 분류를 위한 CNN 네트워크 설계>
2. Conv2D 레이어 (16개 필터, 커널 크기 3x3, 활성화 함수 'relu'): 입력 이미지로부터 특징 추출, 16개의 필터 사용, ReLU로 비선형성 추가
3. MaxPool2D 레이어 (풀 크기 2x2): 특징 맵을 다운샘플링한다. 2x2 풀을 사용하여 특징 맵을 반으로 줄인다
4. Conv2D 레이어 (32개 필터, 커널 크기 3x3, 활성화 함수 'relu'): 더 많은 특징을 추출하기 위해 32개의 필터 사용
6. Flatten 레이어: 다차원 특징 맵 -> 벡터로 변환
7. Dense 레이어: 완전 연결된 은닉층, 32개의 유닛을 가지며 이전 레이어에서 받은 특징을 고차원 공간으로 매핑
8. Dense 레이어: 출력 레이어로 각 클래스에 대한 확률 출력 by softmax 함수, MNIST 데이터셋의 10가지 클래스(0부터 9까지)에 대응
"""

# %%
model.summary()

# 5) 네트워크 학습시키기
print("Before Reshape - x_train_norm shape: {}".format(x_train_norm.shape))
print("Before Reshape - x_test_norm shape: {}".format(x_test_norm.shape))

# -1을 써서 차원의 크기를 자동으로 계산
#x_train_norm 배열의 총 요소 수를 유지하면서 나머지를 28x28x1(28x28 크기의 이미지가 1개의 채널을 갖는 형태)로 재구성
x_train_reshaped=x_train_norm.reshape( -1, 28, 28, 1)
x_test_reshaped=x_test_norm.reshape( -1, 28, 28, 1)

print("After Reshape - x_train_reshaped shape: {}".format(x_train_reshaped.shape))
print("After Reshape - x_test_reshaped shape: {}".format(x_test_reshaped.shape))

# 6) 모델 학습시키기
model.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
model.fit(x_train_reshaped, y_train, epochs=10)
"""
<모델 학습>
1. 신경망 모델의 학습 과정을 설정(compile): adam 최적화 알고리즘 사용 + 희소한 label을 가진 분류 모델에 적합한 손실함수 + 정확도로 모델을 평가
2. 모델을 실제로 학습함: x_train_reshaped는 학습 data-set로 image 데이터가 포함됨, y_train은 학습 date-set의 label로 각 이미지에 대한 정답 레이블이 포함됨, data-set을 10번 반복해서 학습
"""

# %%
# 7) 테스트 데이터로 정확도 확인
test_loss, test_accuracy = model.evaluate(x_test_reshaped, y_test, verbose=2)
print("test_loss: {} ".format(test_loss))
print("test_accuracy: {}".format(test_accuracy))
"""
1. 모델 평가: verbose=2는 평가 과정의 로그를 얼마나 자세히 출력할지 결정
2/3. 테스트 손실/정확도 출력
출력 결과: 313/313 - 1s - loss: 0.0316 - accuracy: 0.9908 - 569ms/epoch - 2ms/step
313/313: 전체 test data의 샘플 수, 1s: 수행된 시간, 2ms/step은 각 스텝당 평균 소요 시간이며 스텝은 모델이 한 번의 배치를 처리하는 과정을 나타냄
"""

# %%
# 8) 예측값, 실제값 확인
predicted_result = model.predict(x_test_reshaped)  # model이 추론한 확률값
predicted_labels = np.argmax(predicted_result, axis=1) #예측한 결과 중 가장 높은 확률을 가진 클래스의 index

idx=0 #1번째 x_test를 보자 
print('model.predict() 결과 : ', predicted_result[idx])
print('model이 추론한 가장 가능성이 높은 결과 : ', predicted_labels[idx])
print('실제 데이터의 라벨 : ', y_test[idx])

# 실제 이미지 확인
plt.imshow(x_test[idx],cmap=plt.cm.binary)
plt.show()


# %%
# 잘못 예측한 예시를 list에 저장
import random
wrong_predict_list=[]
for i, _ in enumerate(predicted_labels):
    # i번째 test_labels과 y_test이 다른 경우만 모아 보자 
    if predicted_labels[i] != y_test[i]:
        wrong_predict_list.append(i)

# wrong_predict_list 에서 랜덤하게 5개의 예시를 선택
samples = random.choices(population=wrong_predict_list, k=5)
# 선택된 예시들에 대해 예측 결과 & 실제 라벨을 출력하고 시각화
for n in samples:
    print("예측확률분포: " + str(predicted_result[n])) #numpy 배열을 문자열로 변환
    print("라벨: " + str(y_test[n]) + ", 예측결과: " + str(predicted_labels[n]))
    plt.imshow(x_test[n], cmap=plt.cm.binary)
    plt.show()

# %%
# 9) 더 좋은 네트워크 생성
# 바꿔 볼 수 있는 하이퍼파라미터들
n_channel_1=32 # 합성곱 층의 필터 개수 (16->32)
n_channel_2=64 # (32->64)
n_dense=64 # dense layer의 뉴런 수 (32->64)
n_train_epoch=20 # (10->20)

model=keras.models.Sequential()
model.add(keras.layers.Conv2D(n_channel_1, (3,3), activation='relu', input_shape=(28,28,1)))
model.add(keras.layers.MaxPool2D(2,2))
model.add(keras.layers.Conv2D(n_channel_2, (3,3), activation='relu'))
model.add(keras.layers.MaxPooling2D((2,2)))
model.add(keras.layers.Flatten())
model.add(keras.layers.Dense(n_dense, activation='relu'))
model.add(keras.layers.Dense(10, activation='softmax'))

model.summary()
model.compile(optimizer='adam',
             loss='sparse_categorical_crossentropy',
             metrics=['accuracy'])

# 모델 훈련
model.fit(x_train_reshaped, y_train, epochs=n_train_epoch)

# 모델 시험
test_loss, test_accuracy = model.evaluate(x_test_reshaped, y_test, verbose=2)
print("test_loss: {} ".format(test_loss))
print("test_accuracy: {}".format(test_accuracy))


